import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;


public class Network {

	public ArrayList<Node> Nodes;
	public ArrayList<Circle> Circles;
	public ArrayList<Integer> ValidCircles;
	public ArrayList<Circle> RetainedCircles;
	public ArrayList<Integer> RetainedValidCircles;
	public EdgeList List;
	public int NodeCount;
	public int FeatureCount;
	public double ThresholdLower;
	private double Lambda;
	
	public Network(int nodeCount, int featureCount, double lambda, double thresholdlower){
		
		this.Nodes = new ArrayList<Node>();
		this.Circles = new ArrayList<Circle>();
		this.ValidCircles = new ArrayList<Integer>();
		this.RetainedCircles = new ArrayList<Circle>();
		this.RetainedValidCircles = new ArrayList<Integer>();
		this.NodeCount = nodeCount;
		this.FeatureCount = featureCount;
		this.Lambda = lambda;
		this.ThresholdLower = thresholdlower;
		
	}
	
	public void InitializeCircles(double thresHold){
		
		int i;		
		
		for(i = 0; i < this.NodeCount; i++){
			Circle circle = new Circle(i, thresHold);
			this.Circles.add(i, circle);
			this.ValidCircles.add(Integer.valueOf(i));
			this.RetainedCircles.add(i, circle);
			this.RetainedValidCircles.add(Integer.valueOf(i));
		}
		for(i = 0; i < this.NodeCount; i++){
			
			this.Circles.get(i).Nodes.add(Integer.valueOf(i));
			this.RetainedCircles.get(i).Nodes.add(Integer.valueOf(i));
		}
		
		
	}
	
	public void UpdateRetainedCircles(){
		
		this.RetainedCircles.clear();
		this.RetainedValidCircles.clear();
		
		int i;
		
		for(i = 0; i < this.Circles.size(); i++){
			Circle dupcircle = this.Circles.get(i).Duplicate();
			this.RetainedCircles.add(dupcircle);
		}
		
		this.RetainedValidCircles = new ArrayList<Integer>(this.ValidCircles);
		
	}
	
	public void InitializeNodes(String fileName){
		
		
		FileReader f1;
		BufferedReader br;
		int i,j;		
		
		for(i = 0; i < this.NodeCount; i++){
			Node node = new Node(i, this.FeatureCount);
			this.Nodes.add(i, node);
		}
		
		try {
			
			f1 = new FileReader(fileName);
			br = new BufferedReader(f1);
					
			for(i = 0; i < this.NodeCount; i++){
				
				this.Nodes.get(i).ID = Integer.parseInt(br.readLine());
				this.Nodes.get(i).ListCircles.add(Integer.valueOf(i));
				for(j = 0; j < this.FeatureCount; j++){
					this.Nodes.get(i).FeatureVector[j] = Double.parseDouble(br.readLine());
				}
				
			}
			
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
		
	}
	
	public void InitializeEdgeList(String fileName){
		
		this.List = new EdgeList(fileName);
		
	}
	
	public void UpdateNodeMembership(int nodeID){
		
		Node node = this.GetNode(nodeID);		
		
		int incirclecount = node.ListCircles.size();
		int notincirclecount = this.ValidCircles.size() - incirclecount;
		ArrayList<Integer> notincircle = this.GetDifferentCircles(this.ValidCircles, node.ListCircles);
		
		int k1, k2, addcircle, removecircle;
		int i;
		
		k1 = (int)(Math.random() * notincirclecount );
		k2 = (int)(Math.random() * incirclecount);
		addcircle = (k1 + incirclecount)/(incirclecount);
		removecircle = (k2 + incirclecount)/(incirclecount);
		
		
		for(i = 0; i < addcircle; i++){
			this.AddNodeToCircle(nodeID, notincircle.get(i));
		}
		for(i = 0; i < removecircle; i++){
			this.RemoveNodeFromCircle(nodeID, node.ListCircles.get(i));
		}
		
		return;
		
	}	
		
	public double GetLogLikelihood(){
		
		double loglikelihood = 0.0;
		double edgeprobpos;
		double edgeprobneg;
		int i,j;
		
		for(i = 0; i < this.NodeCount; i++){
			for(j = i+1; j < this.NodeCount; j++){
				 edgeprobpos = this.ComputeEdgeProbabilityPos(i, j);
				 edgeprobneg = 1.0 - edgeprobpos;
				if(this.List.isEdge(i, j)){
					loglikelihood = loglikelihood + Math.log(edgeprobpos);
				}
				else{
					loglikelihood = loglikelihood + Math.log(edgeprobneg);					
				}
			}
		}
		return loglikelihood;
	}
	
	public void UpdateCircleThresholds(){
		
		int i, circleID;
		ArrayList<Integer> circlesfordeletion = new ArrayList<Integer>();
		for(i = 0; i < this.ValidCircles.size(); i++){
			
			circleID = this.ValidCircles.get(i);
			this.UpdateCircleThreshold(circleID);
			if(this.GetCircle(circleID).Threshold < this.ThresholdLower){
				circlesfordeletion.add(Integer.valueOf(circleID));
			}
		}
		
		for(i = 0; i < circlesfordeletion.size(); i++){
			this.DeleteCircle(circlesfordeletion.get(i));
		}
	}
	
	public void OverWriteCircles(String fileName){
		
		int i,j;
		int circleID;
		FileWriter f;
		BufferedWriter bw;
		try {
			
			f = new FileWriter(fileName);
			bw = new BufferedWriter(f);
			for(i = 0; i < this.ValidCircles.size(); i++){
				
				circleID = this.ValidCircles.get(i);
				Circle circle = this.GetCircle(circleID);
				bw.write("#Circle "+circleID + ":\n");
				for(j = 0; j < circle.Nodes.size(); j++){
					bw.write(circle.Nodes.get(j)+" ");
				}
				bw.write("\n");
			}
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return;
	}
	
	private void UpdateCircleThreshold(int circleID){
		
		Circle circle = this.GetCircle(circleID);
		int circleindex = this.GetCircleIndex(circleID);
		int i;
		double threshold = 100.0;
		double dist;
		for(i = 0; i < circle.Nodes.size(); i++){
			
			dist = this.ComputeNodeCircleDistance(circleID, circle.Nodes.get(i)); 
			if(dist < threshold){
				threshold = dist; 
			}
		}
		
		this.Circles.get(circleindex).Threshold = threshold;		
		
	}
	
	  
	
	private void DeleteCircle(int circleID){
		
		int i, index;
		Circle circle = this.GetCircle(circleID);
		for(i = 0; i < circle.Nodes.size(); i++){
			 this.RemoveNodeFromCircle(circle.Nodes.get(i), circleID);			
		}
		
		index = this.ValidCircles.indexOf(Integer.valueOf(circleID));
		if(index == -1){
			System.out.println("Invalid Remove Circle Request : Circle:"+ circleID);
		}
		this.ValidCircles.remove(index);
		
	}
	
	private void AddNodeToCircle(int nodeID, int circleID){
		
		int nodeIndex = this.GetNodeIndex(nodeID);
		int circleIndex = this.GetCircleIndex(circleID);
		
		if(nodeIndex == -1 || circleIndex == -1){
			try {
				throw new Exception("Invalid Membership addition operation!!!");
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		this.Nodes.get(nodeIndex).AddToCircle(circleID);
		this.Circles.get(circleIndex).AddNode(nodeID);
		
	}
	
	private void RemoveNodeFromCircle(int nodeID, int circleID){
		
		int nodeIndex = this.GetNodeIndex(nodeID);
		int circleIndex = this.GetCircleIndex(circleID);
		
		if(nodeIndex == -1 || circleIndex == -1){
			try {
				throw new Exception("Invalid Membership addition operation!!!");
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		this.Nodes.get(nodeIndex).DeleteCircle(circleID);
		this.Circles.get(circleIndex).RemoveNode(nodeID);
		
	}
	
	private double ComputeNodeCircleDistance(int circleID, int nodeID){
		
		int i;
		double distTemp = 0.0;
		double distance = 0.0;
		
		Circle circle = GetCircle(circleID);
		Node node = GetNode(nodeID);
		
		for(i = 0; i < circle.Nodes.size(); i++){
			distTemp = node.ComputeDistance(GetNode(circle.Nodes.get(i)));
			distance = distance + distTemp;
		}
		if(circle.Nodes.size() == 0){
			return 0.0;
		}
		return (distance / (double)circle.Nodes.size());
		
	}
	
	private ArrayList<Integer> GetCommonCircles(ArrayList<Integer> A1, ArrayList<Integer> A2){
		
		ArrayList<Integer> intersection = new ArrayList<Integer>(A1);
		intersection.retainAll(A2);
		return intersection;
	}
	
	private ArrayList<Integer> GetDifferentCircles(ArrayList<Integer> A1, ArrayList<Integer> A2){
		
		ArrayList<Integer> intersection = new ArrayList<Integer>(A1);
		intersection.removeAll(A2);
		return intersection;
	}
	
	private double computeBeta1(int nodeID1, int nodeID2){
		
		double beta1 = 0.0;
		int i;
		Node node1 = this.GetNode(nodeID1);
		Node node2 = this.GetNode(nodeID2);
		double sim = node1.ComputeSim(node2);
		
		ArrayList<Integer> commoncircles = this.GetCommonCircles(node1.ListCircles, node2.ListCircles);
		for(i = 0; i < commoncircles.size(); i++){
			
			Circle circle = this.GetCircle(commoncircles.get(i));
			beta1 = beta1 + 1.0/(sim - circle.Threshold + this.Lambda);			
		}
		
		return beta1;
		
	}
	
	private double computeBeta2(int nodeID1, int nodeID2){
		
		double beta2 = 0.0;
		int i;
		Node node1 = this.GetNode(nodeID1);
		Node node2 = this.GetNode(nodeID2);
		double sim = node1.ComputeSim(node2);
		
		ArrayList<Integer> commoncircles = this.GetCommonCircles(node1.ListCircles, node2.ListCircles);
		ArrayList<Integer> diffcircles = this.GetDifferentCircles(this.ValidCircles, commoncircles);
		for(i = 0; i < diffcircles.size(); i++){
			
			Circle circle = this.GetCircle(diffcircles.get(i));
			beta2 = beta2 + 1.0/(sim - circle.Threshold + this.Lambda);			
		}
		
		return beta2;
		
	}
	
	private double ComputeBeta(int nodeID1, int nodeID2){
		
		double beta = 0.0;
		double beta1 = this.computeBeta1(nodeID1, nodeID2);
		double beta2 = this.computeBeta2(nodeID1, nodeID2);
		beta = Math.exp((beta1*beta1) - (beta2*beta2));
		return beta;
	}
	
	private double ComputeEdgeProbabilityPos(int nodeID1, int nodeID2){
		
		double prob = 0.0;
		double beta = this.ComputeBeta(nodeID1, nodeID2);
		prob = beta / (1.0 + beta);
		return prob;
	}
	
	

	private Node GetNode(int nodeID) {
		// TODO Auto-generated method stub
		int i;
		Node node = null;
		for(i = 0; i < this.Nodes.size(); i++){
			if(this.Nodes.get(i).ID == nodeID){
				node = this.Nodes.get(i);
			}
		}
		if(node == null){
			System.out.println("Node not found");
		}
		return node;
		
	}

	private Circle GetCircle(int circleID) {
		// TODO Auto-generated method stub
		int i;
		Circle circle = null;
		for(i = 0; i < this.Circles.size(); i++){
			if(this.Circles.get(i).ID == circleID){
				circle = this.Circles.get(i);
			}
		}
		if(circle == null){
			System.out.println("Circle not found");
		}
		return circle;
	}
	
	private int GetNodeIndex(int nodeID) {
		// TODO Auto-generated method stub
		int i, index;
		index = -1;
		for(i = 0; i < this.Nodes.size(); i++){
			if(this.Nodes.get(i).ID == nodeID){
				index = i;
			}
		}
		if(index == -1){
			System.out.println("Node not found");
		}
		return index;
		
	}

	private int GetCircleIndex(int circleID) {
		// TODO Auto-generated method stub
		int i, index;
		index = -1;
		for(i = 0; i < this.Circles.size(); i++){
			if(this.Circles.get(i).ID == circleID){
				index = i;
			}
		}
		if(index == -1){
			System.out.println("Node not found");
		}
		return index;
	}
	

}
